import pytest  # type: ignore

from eventcounter import EventCounter


########################################################
#
# Test Plan
#
########################################################

# 1) Create instance and log events
# 2) Create 2 instances, log events and merge
# 3) Create an instance and get categories


_CATEGORIES: list[str] = ["one", "two", "three", "four"]


@pytest.fixture
def test_source_categories_0() -> list[str]:
    return _CATEGORIES


def mk_eventcounter_one(name: str, cats: list[str]) -> EventCounter:
    ec: EventCounter = EventCounter(name)
    for c in cats:
        ec.log(c)
    return ec


def mk_eventcounter_linear(name: str, cats: list[str]) -> EventCounter:
    ec: EventCounter = EventCounter(name)
    for i in range(len(cats)):
        ec.log(cats[i], i + 1)
    return ec


def test_1_create(test_source_categories_0: list[str]) -> None:
    """Test __init__(), log(), get_categories()"""
    cats: list[str] = test_source_categories_0
    ec: EventCounter = mk_eventcounter_one("Test", cats)
    assert ec.categories.sort() == cats.sort(), "Failed to create categories"
    assert ec.sum(cats) == len(cats), "Incorrect event count"
    assert ec.get(cats[1]) == 1, "Incorrect value"


def test_2_create(test_source_categories_0: list[str]) -> None:
    cats: list[str] = test_source_categories_0
    ec: EventCounter = mk_eventcounter_linear("Test", cats)
    assert ec.categories.sort() == cats.sort(), "Failed to create categories"
    assert ec.sum(cats) == (1 + len(cats)) / 2 * len(cats), "Incorrect event count"
    assert ec.get(cats[1]) == 2, "Incorrect value"


def test_3_merge(test_source_categories_0: list[str]) -> None:
    cats: list[str] = test_source_categories_0
    ec1: EventCounter = mk_eventcounter_linear("Test 1", cats)
    ec2: EventCounter = mk_eventcounter_linear("Test 2", cats)
    ec1.merge(ec2)
    assert ec1.categories.sort() == cats.sort(), "Merging categories failed categories"
    assert ec1.sum(cats) == 2 * (1 + len(cats)) / 2 * len(cats), "Incorrect event count"


def test_4_merge_child(test_source_categories_0: list[str]) -> None:
    cats: list[str] = test_source_categories_0
    ec1: EventCounter = mk_eventcounter_linear("Test 1", cats)
    ec2: EventCounter = mk_eventcounter_linear("Test 2", cats)
    ec1.merge_child(ec2)
    assert len(ec1.categories) == 2 * len(cats), "Incorrect category count"
    assert ec1.sum(ec1.categories) == 2 * (1 + len(cats)) / 2 * len(cats), (
        "Incorrect event count"
    )


def test_5_error_categories(test_source_categories_0: list[str]) -> None:
    cats: list[str] = test_source_categories_0
    ec: EventCounter = mk_eventcounter_linear("Test", cats)
    assert not ec.error_status, "Unintended errors logged"
    # assert ec.add_error_categories(cats[:2]), "Errors not logged"


def test_6_categories(test_source_categories_0: list[str], capsys) -> None:
    """Test categories methods"""
    cats: list[str] = test_source_categories_0
    ec: EventCounter = mk_eventcounter_linear("Test", cats)
    assert isinstance(ec._get_str(cats[1]), str), (
        f"Incorrect type returned: {type(ec._get_str(cats[1]))}"
    )
    assert isinstance(ec.get_long_cat(cats[1]), str), (
        f"Incorrect type returned: {type(ec.get_long_cat(cats[1]))}"
    )
    assert isinstance(ec.values, dict), f"Incorrect type returned: {type(ec.values)}"
    assert isinstance(ec.print(do_print=False), str), "Incorrect type returned"
    ec.print()
    captured = capsys.readouterr()
    assert len(captured.out.splitlines()) == 4, (
        f"Incorrect output: {len(captured.out.splitlines())}"
    )



## -----------------------------------------------------------
#### Class EventCounter()
# Class to log/count events, pass them to parent function
# and merge the results
## -----------------------------------------------------------

from collections import defaultdict
from typing import Optional, Iterable, Callable
from asyncio import gather, Task
from deprecated import deprecated
from multilevellogger import getMultiLevelLogger, MultiLevelLogger

logger: MultiLevelLogger = getMultiLevelLogger(__name__)

debug = logger.debug
message = logger.message
verbose = logger.verbose
error = logger.error

# FuncTypeFormatter = Callable[[str], str]
# FuncTypeFormatterParam = Optional[FuncTypeFormatter]


class EventCounter:
    """
    Count events and merge event counts from other EventCounters as a child or sibling

    Usage:

    from eventcounter import EventCounter

    stats = EventCounter("Demo counter")

    stats.log("tests")
    stats.log("errors")
    stats.log("added", 3)

    stats.print()
    """

    def __init__(
        self,
        name: str = "",
        totals: Optional[str] = None,
        # categories: set[str] = set(),
        errors: set[str] = set(),
        separator: str = ": ",
        format_spec: str = "{category:<{col_width}}{separator}{value}",
        print_func: Callable[[str], None] = print,
    ):
        assert name is not None, "param 'name' cannot be None"
        # assert categories is not None, "param 'categories' cannot be None"
        assert errors is not None, "param 'errors' cannot be None"

        self.name: str = name
        self._log: defaultdict[str, int] = defaultdict(
            int
        )  # returns 0 if key not found
        self._errors: set[str] = errors
        self._totals = totals
        self._separator: str = separator
        self._format_spec: str = format_spec
        self._print_func: Callable[[str], None] = print_func

    def log(self, category: str, count: int = 1) -> int:
        assert category is not None, "category cannot be None"
        assert count is not None, "count cannot be None"
        self._log[category] += count
        return self._log[category]

    def get_long_cat(self, category: str) -> str:
        """
        Get the long category name including the name of the EventCounter
        """
        assert category is not None, "param 'category' cannot be None"
        return f"{self.name}{self._separator}{category}"

    def _get_str(self, category: str, col_width: int = 40) -> str:
        assert category is not None, "category cannot be None"
        return self._format_spec.format(
            category=category,
            col_width=col_width,
            separator=self._separator,
            value=self.get(category),
        )

    # TODO: remove later
    @deprecated(version="0.4.0", reason="use get() instead")
    def get_value(self, category: str) -> int:
        assert category is not None, "param 'category' cannot be None"
        return self._log[category]

    def get(self, category: str) -> int:
        """ "
        Get the value of a category
        """
        assert category is not None, "param 'category' cannot be None"
        return self._log[category]

    # TODO: remove later
    @deprecated(version="0.4.0", reason="use values() property instead")
    def get_values(self) -> dict[str, int]:
        return self._log

    @property
    def values(self) -> dict[str, int]:
        return self._log.copy()

    @property
    def error_values(self) -> dict[str, int]:
        """
        Get the values of the error categories"""
        return {key: self._log[key] for key in self._log.keys() & self._errors}

    def sum(self, categories: Iterable[str]) -> int:
        """
        Sum the values of the categories
        """
        ret = 0
        for cat in categories:
            ret += self.get(cat)
        return ret

    # TODO: remove later
    @deprecated(version="0.4.0", reason="use categories() property instead")
    def get_categories(self) -> list[str]:
        return list(self._log.keys())

    @property
    def categories(self) -> list[str]:
        return sorted(self._log.keys())

    @property
    def errors(self) -> list[str]:
        return sorted(self._errors)

    # TODO: remove later
    @deprecated(version="0.4.0", reason="use error_status() property instead")
    def get_error_status(self) -> bool:
        return self.sum(self._errors) > 0

    @property
    def error_status(self) -> bool:
        return self.sum(self._errors) > 0

    def merge(self, B: "EventCounter") -> bool:
        """Merge two EventCounter instances together"""
        assert isinstance(B, EventCounter), (
            f"input is not type of 'EventCounter' but: {type(B)}"
        )

        try:
            for cat in B.categories:
                value: int = B.get(cat)
                self.log(cat, value)
                if self._totals is not None:
                    # self.log(f"{self._totals}: {cat}", value)
                    self.log(self._totals, value)
            return True
        except Exception as err:
            error(f"{err}")
        return False

    def merge_child(self, B: "EventCounter") -> bool:
        """Merge two EventCounter instances together"""
        assert isinstance(B, EventCounter), (
            f"input is not type of 'EventCounter' but: {type(B)}"
        )

        try:
            for cat in B.categories:
                value: int = B.get(cat)
                self.log(B.get_long_cat(cat), value)
                if self._totals is not None:
                    self.log(f"{self._totals}: {cat}", value)
            # self._error_status = self._error_status or B.get_error_status()
            return True
        except Exception as err:
            error(f"{err}")
        return False

    def get_header(self) -> str:
        return f"{self.name}" + (
            f"{self._separator}ERROR occurred" if self.error_status else ""
        )

    @property
    def header(self) -> str:
        return f"{self.name}" + (
            f"{self._separator}ERROR occurred" if self.error_status else ""
        )

    @property
    def col_width(self) -> int:
        """
        Set category column width based on the longest column
        """
        return len(max(self._log.keys(), key=len)) + 3

    def print(self, do_print: bool = True, clean: bool = False) -> Optional[str]:
        try:
            width: int = self.col_width
            if do_print:
                message(self.get_header())
                for cat in sorted(self._log):
                    if clean and self.get(cat) == 0:
                        continue
                    self._print_func(self._get_str(cat, width))
                return None
            else:
                ret = self.get_header()
                for cat in sorted(self._log):
                    if clean and self.get_value(cat) == 0:
                        continue
                    ret += f"\n{self._get_str(cat, width)}"
                return ret
        except Exception as err:
            error(f"{err}")
        return None

    @deprecated(version="0.4.0", reason="use gather() instead")
    async def gather_stats(
        self, tasks: list[Task], merge_child: bool = True, cancel: bool = True
    ) -> None:
        """
        Wrapper to gather results from tasks and return the stats and the LAST exception
        """
        await self.gather(tasks, merge_child, cancel)

    async def gather(
        self, tasks: list[Task], merge_child: bool = True, cancel: bool = True
    ) -> None:
        """
        Wrapper to gather results from tasks and return the stats and the LAST exception
        """
        if cancel:
            for task in tasks:
                task.cancel()
        for res in await gather(*tasks, return_exceptions=True):
            if isinstance(res, EventCounter):
                if merge_child:
                    self.merge_child(res)
                else:
                    self.merge(res)
            elif type(res) is BaseException:
                error(f"Task raised an exception: {res}")
        return None



from eventcounter import EventCounter
from random import randint

counter: EventCounter = EventCounter("Numbers")

for _ in range(100):
    if randint(1, 10) % 2 == 0:
        counter.log("even number")
    else:
        counter.log("odd number")

counter.print()
